#!/bin/sh

# prerequisites on RHEL 7
#  yum install gcc-c++
# prerequisites on Debian/Ubuntu
#  apt-get install g++

# Stripped path. Can not use users path or else resulting
# SDK may not work for other users.
PATH=/usr/local/bin:/usr/bin:/usr/sbin:/bin:/sbin

# Slow down the builds to make troubleshooting easier.
MAKE_CONCURRENCY=1

MAKE_ARCH_LIST="x64 x86"
# arm ppc64 ppc

# Support for building binary kernel modules for Redhat based
# Linux distros.
# MAKE_RHEL5_KERNELS=1
MAKE_RHEL6_KERNELS=1
# MAKE_RHEL7_KERNELS=1

# Changing versions of many of these packages will
# require script changes and patching to resolve
# configure/make failures. Remember that most of
# these packages are built just to get a backward
# compatible *.so symbol table to link against, the
# compiled binaries never actually execute.

config_common()
{
   echo config_common

   unset binutils_name binutils_package binutils_url
   unset gcc_name gcc_package gcc_url

   atk_ver_=1.12 # RHEL 5
   atk_ver=$atk_ver_.4
   bison_ver=3.0.4
   binutils_ver=2.21.1a
   byacc_ver=
   bzip2_ver=1.0.6
   cairo_ver=1.8.10 # RHEL 5 is 1.2.4
   # expat_ver=2.0.1
   findutils_ver=4.4.2
   flex_ver=2.6.0
   fontconfig_ver=2.7.3 # RHEL 5 is 2.4.1
   freetype_ver=2.4.6 # RHEL 5 is 2.2.1
   gawk_ver=3.1.8
   gcc_build_ver=5.3.0
   gcc_pass1_ver=4.4.5
   gcc_pass2_ver=4.4.5
   gcc_final_ver=5.3.0
   gettext_ver=0.18.1 # RHEL 5 is 0.14.6
   glib_ver_=2.29 # RHEL 5 is 2.12 but older versions cause ripple of cross compile issues
   glib_ver=$glib_ver_.4
   glibc_ver=2.6.1 # RHEL 5 is glibc 2.5, 2.3 shipped in Debian 2007, but 2.6 is mostly ABI compat and cross compiles better
   gmp_ver=4.3.2
   gperf_ver=3.0.4
   groff_ver=1.20.1 # 1.21 fails build of .ps files in libX11
   gtk_ver_=2.10 # RHEL 5
   gtk_ver=$gtk_ver_.14
   inputproto_ver=2.0
   kbproto_ver=1.0.4
   libarchive_ver=2.8.4
   libelf_ver=0.8.13
   libffi_ver=3.0.10
   libgcrypt_ver=1.4.6
   libgpg_error_ver=1.10
   libiconv_ver=1.14
   libpng_url_ver=12
   libpng_ver=1.2.46 # RHEL 5 is 1.2.10
   libpthread_stubs_ver=0.3
   libtool_ver=2.4
   libx11_ver=1.3.2 # from X11R7.5, X11 has good forward/backward ABI/API compat
   libxau_ver=1.0.5
   libxcb_ver=1.7
   libxext_ver=1.1.1
   libxml2_ver=2.7.8
   libxrender_ver=0.9.5
   libxslt_ver=1.1.26
   m4_ver=1.4.16
   make_ver=3.81
   mpc_ver=0.8.2
   mpfr_ver=2.4.2
   ncurses_ver=5.5 # RHEL 5 is 5.5
   renderproto_ver=0.11
   rhel5_ver_=2.6.18
   rhel5_ver=$rhel5_ver_-x.el5
   rhel6_ver_=2.6.32
   rhel6_ver=$rhel6_ver_-x.el6
   rhel7_ver_=3.10.0
   rhel7_ver=$rhel7_ver_-x.el7
   pango_ver_=1.28 # RHEL 5 is 1.14.9
   pango_ver=$pango_ver_.4
   pixman_ver=0.20.2
   pkg_config_ver=0.25 # 0.26 has circular glib dependency
   python_ver=2.7.1
   sed_ver=4.2.1
   # wget_ver=1.12
   x11_ver=X11R7.5
   xcb_proto_ver=1.6
   xextproto_ver=7.1.1
   xproto_ver=7.0.16
   xtrans_ver=1.2.6
   xz_ver=5.1.4beta
   zlib_ver=1.2.5

   # gnumirror_url="http://ftpmirror.gnu.org"
   gnumirror_url="http://mirrors.kernel.org/gnu"
   sourceforgemirror_url="http://downloads.sourceforge.net/project"

   atk_url_="http://ftp.gnome.org/pub/gnome/sources/atk/$atk_ver_/"
   binutils_suffix=.tar.bz2
   bison_package="bison-$bison_ver.tar.gz"
   byacc_url="http://invisible-island.net/datafiles/release/byacc.tar.gz"
   bzip2_url_="http://www.bzip.org/$bzip2_ver/"
   cairo_url_="http://cairographics.org/releases/"
   # expat_url_="$sourceforgemirror_url/expat/expat/$expat_ver/"
   flex_url_="$sourceforgemirror_url/flex/"
   fontconfig_url_="http://www.freedesktop.org/software/fontconfig/release/"
   freetype_url_="http://download.savannah.gnu.org/releases/freetype/"
   gcc_url_="$gnumirror_url/gcc/gcc-$gcc_ver/"
   glibc_ports_name="glibc-ports-$glibc_ver"
   glibc_ports_url_="$gnumirror_url/glibc/"
   glib_url_="http://ftp.gnome.org/pub/gnome/sources/glib/$glib_ver_/"
   gtk_name="gtk+-$gtk_ver"
   gtk_url_="http://ftp.gnome.org/pub/gnome/sources/gtk+/$gtk_ver_/"
   libarchive_url_="http://libarchive.googlecode.com/files/"
   libelf_url_="http://www.mr511.de/software/"
   libffi_url_="ftp://sourceware.org/pub/libffi/"
   libgcrypt_url_="ftp://ftp.gnupg.org/gcrypt/libgcrypt/"
   libgpg_error_name="libgpg-error-$libgpg_error_ver"
   libgpg_error_package="libgpg-error-$libgpg_error_ver.tar.bz2"
   libgpg_error_url_="ftp://ftp.gnupg.org/gcrypt/libgpg-error/"
   # libpng_url_="$sourceforgemirror_url/libpng/libpng$libpng_url_ver/older-releases/$libpng_ver/"
   libpng_url_="ftp://ftp.simplesystems.org/pub/png/src/history/libpng12/"
   libxml2_url_="ftp://xmlsoft.org/libxslt/"
   libxslt_url_="ftp://xmlsoft.org/libxslt/"
   mpc_url_="http://www.multiprecision.org/mpc/download/"
   pango_url_="http://ftp.gnome.org/pub/gnome/sources/pango/$pango_ver_/"
   pixman_url_="http://cairographics.org/releases/"
   pkg_config_name="pkg-config-$pkg_config_ver"
   pkg_config_url_="http://pkgconfig.freedesktop.org/releases/"
   python_name="Python-$python_ver"
   python_package="$python_name.tgz"
   python_url_="http://www.python.org/ftp/python/$python_ver/"
   rhel5_name="linux-$rhel5_ver"
   rhel5_package="$rhel5_name.20111020.tar.gz"
   rhel5_url_="http://www.turtlewar.org/resources/"
   rhel6_name="linux-$rhel6_ver"
   rhel6_package="$rhel6_name.20111101.tar.gz"
   rhel6_url_="http://www.turtlewar.org/resources/"
   rhel7_name="linux-$rhel7_ver"
   rhel7_package="$rhel7_name.20141107.tar.gz"
   rhel7_url_="http://www.turtlewar.org/resources/"
   xz_url_="http://tukaani.org/xz/"
   #zlib_url_="$sourceforgemirror_url/libpng/zlib/$zlib_ver/"
   zlib_url_="zlib.net/fossils/"
   linux_ver="$rhel6_ver"
   linux_name="$rhel6_name"
   linux_package="$rhel6_package"
   linux_url_="$rhel6_url_"

   xorgmirror_url="http://www.x.org/releases"
   # xorgmirror_url="http://ftp.vim.org/ftp/ftp/windowing/X/releases"

   inputproto_url_="$xorgmirror_url/$x11_ver/src/proto/"
   kbproto_url_="$xorgmirror_url/$x11_ver/src/proto/"
   libpthread_stubs_name="libpthread-stubs-$libpthread_stubs_ver"
   libpthread_stubs_url_="http://xcb.freedesktop.org/dist/"
   libx11_name="libX11-$libx11_ver"
   libx11_url_="$xorgmirror_url/$x11_ver/src/lib/"
   libxau_name="libXau-$libxau_ver"
   libxau_url_="$xorgmirror_url/$x11_ver/src/lib/"
   libxcb_url_="http://xcb.freedesktop.org/dist/"
   libxext_name="libXext-$libxext_ver"
   libxext_url_="$xorgmirror_url/$x11_ver/src/lib/"
   libxrender_name="libXrender-$libxrender_ver"
   libxrender_url_="$xorgmirror_url/$x11_ver/src/lib/"
   renderproto_url_="$xorgmirror_url/$x11_ver/src/proto/"
   xcb_proto_name="xcb-proto-$xcb_proto_ver"
   xcb_proto_url_="http://xcb.freedesktop.org/dist/"
   xextproto_url_="$xorgmirror_url/$x11_ver/src/proto/"
   xproto_url_="$xorgmirror_url/$x11_ver/src/proto/"
   xtrans_url_="ftp://ftp.x.org/pub/individual/lib/"
}

config_build()
{
   echo config_build

   unset binutils_name binutils_package binutils_url
   binutils_ver=2.26
   binutils_package="binutils-$binutils_ver.tar.gz"
   unset gcc_name gcc_package gcc_url
   config_gcc build

   gcc_ver=$gcc_final_ver

   unset gcc_name gcc_package gcc_url
   gcc_ver=$gcc_libc_ver
   gcc_url_="$gnumirror_url/gcc/gcc-$gcc_ver/"
}

config_x64()
{
   echo config_x64
}

config_x86()
{
   echo config_x86
}

config_arm()
{
   echo config_arm
}

config_ppc64()
{
   echo config_ppc64
}

config_ppc()
{
   echo config_ppc
}

is_arch_enabled()
{
   local platform=$1
   local check
   local res=""

   for check in $MAKE_ARCH_LIST; do
      test $platform = $check && return 0
   done
   return 1
}

setup_env()
{
   # platform
   # tuple
   # common_cflags
   # build_cflags
   # build_ldflags
   # build_tuple
   local default="$1"
   local prefix

   if [ $tuple = build ]; then
      target_cflags="$build_cflags"
      target_ldflags="$build_ldflags"
   else
      target_cflags="-isystem $ptlinsdk/$platform/include"
      target_ldflags="-L$ptlinsdk/$platform/usr/lib"
   fi
   prefix="$ptlinsdk/build/bin/$build_tuple"
   export          BUILD_CC="$prefix-gcc"
   export      CC_FOR_BUILD="$prefix-gcc"
   export     CXX_FOR_BUILD="$prefix-g++"
   export     GCC_FOR_BUILD="$prefix-gcc"
   export      AR_FOR_BUILD="$prefix-ar"
   export      AS_FOR_BUILD="$prefix-as"
   export      LD_FOR_BUILD="$prefix-ld"
   export      NM_FOR_BUILD="$prefix-nm"
   export OBJDUMP_FOR_BUILD="$prefix-objdump"
   export  RANLIB_FOR_BUILD="$prefix-ranlib"
   export   STRIP_FOR_BUILD="$prefix-strip"
   export LIBTOOL_FOR_BUILD="$prefix-libtool"
   prefix="$ptlinsdk/build/bin/$tuple"
   export      CC_FOR_HOST="$prefix-gcc"
   export     CXX_FOR_HOST="$prefix-g++"
   export     GCC_FOR_HOST="$prefix-gcc"
   export      AR_FOR_HOST="$prefix-ar"
   export      AS_FOR_HOST="$prefix-as"
   export      LD_FOR_HOST="$prefix-ld"
   export      NM_FOR_HOST="$prefix-nm"
   export OBJDUMP_FOR_HOST="$prefix-objdump"
   export  RANLIB_FOR_HOST="$prefix-ranlib"
   export   STRIP_FOR_HOST="$prefix-strip"
   export LIBTOOL_FOR_HOST="$prefix-libtool"
   export      CC_FOR_TARGET="$prefix-gcc"
   export     CXX_FOR_TARGET="$prefix-g++"
   export     GCC_FOR_TARGET="$prefix-gcc"
   export      AR_FOR_TARGET="$prefix-ar"
   export      AS_FOR_TARGET="$prefix-as"
   export      LD_FOR_TARGET="$prefix-ld"
   export      NM_FOR_TARGET="$prefix-nm"
   export OBJDUMP_FOR_TARGET="$prefix-objdump"
   export  RANLIB_FOR_TARGET="$prefix-ranlib"
   export   STRIP_FOR_TARGET="$prefix-strip"
   export LIBTOOL_FOR_TARGET="$prefix-libtool"
   export CPPFLAGS_FOR_BUILD="-isystem $ptlinsdk/build/usr/include"
   unset LDFLAGS_FOR_BUILD
   export CFLAGS_FOR_BUILD="$common_cflags $CPPFLAGS_FOR_BUILD"
   export CXXFLAGS_FOR_BUILD="$CFLAGS_FOR_BUILD"
   export LIBCFLAGS_FOR_BUILD="$CFLAGS_FOR_BUILD"
   export LIBCXXFLAGS_FOR_BUILD="$CFLAGS_FOR_BUILD"
   unset CPPFLAGS_FOR_HOST
   unset LDFLAGS_FOR_HOST
   export CFLAGS_FOR_HOST="$common_cflags"
   export CXXFLAGS_FOR_HOST="$CFLAGS_FOR_HOST"
   export LIBCFLAGS_FOR_HOST="$CFLAGS_FOR_HOST"
   export LIBCXXFLAGS_FOR_HOST="$CFLAGS_FOR_HOST"
   export CPPFLAGS_FOR_TARGET="$CPPFLAGS_FOR_HOST"
   export LDFLAGS_FOR_TARGET="$LDFLAGS_FOR_HOST"
   export CFLAGS_FOR_TARGET="$CFLAGS_FOR_HOST"
   export CXXFLAGS_FOR_TARGET="$CFLAGS_FOR_TARGET"
   export LIBCFLAGS_FOR_TARGET="$CFLAGS_FOR_TARGET"
   export LIBCXXFLAGS_FOR_TARGET="$CFLAGS_FOR_TARGET"
   unset LDFLAGS
   if [ "$default" = "$tuple" ]; then
      export      CC="$CC_FOR_HOST"
      export     CXX="$CXX_FOR_HOST"
      export     GCC="$GCC_FOR_HOST"
      export      AR="$AR_FOR_HOST"
      export      AS="$AS_FOR_HOST"
      export      LD="$LD_FOR_HOST"
      export      NM="$NM_FOR_HOST"
      export OBJDUMP="$OBJDUMP_FOR_HOST"
      export  RANLIB="$RANLIB_FOR_HOST"
      export   STRIP="$STRIP_FOR_HOST"
      export LIBTOOL="$LIBTOOL_FOR_HOST"
      unset CPPFLAGS
      # export CPPFLAGS="$CPPFLAGS_FOR_HOST"
   elif [ "$default" = "$build_tuple" ] || [ $tuple = $build_tuple ]; then
      export      CC="$CC_FOR_BUILD"
      export     CXX="$CXX_FOR_BUILD"
      export     GCC="$GCC_FOR_BUILD"
      export      AR="$AR_FOR_BUILD"
      export      AS="$AS_FOR_BUILD"
      export      LD="$LD_FOR_BUILD"
      export      NM="$NM_FOR_BUILD"
      export OBJDUMP="$OBJDUMP_FOR_BUILD"
      export  RANLIB="$RANLIB_FOR_BUILD"
      export   STRIP="$STRIP_FOR_BUILD"
      export LIBTOOL="$LIBTOOL_FOR_BUILD"
      export CPPFLAGS="$CPPFLAGS_FOR_BUILD"
      default="$build_tuple"
   else
      unset      CC
      unset     CXX
      unset     GCC
      unset      AR
      unset      AS
      unset      LD
      unset      NM
      unset OBJDUMP
      unset  RANLIB
      unset   STRIP
      unset LIBTOOL
      unset CPPFLAGS
      default="$build_tuple"
   fi
   export CFLAGS="$common_cflags $CPPFLAGS"
   export CXXFLAGS="$CFLAGS"
   export LIBCFLAGS="$CFLAGS"
   export LIBCXXFLAGS="$CFLAGS"
   export PATH="$ptlinsdk/build/bin:$ptlinsdk/build/bin/$default:$ptlinsdk/build/bin/stop:$sys_path"
   unset LD_LIBRARY_PATH
   unset C_INCLUDE_PATH
   unset LIBPATH
   unset LIBRARY_PATH
   unset LD_RUN_PATH
}

echo_env()
{
   env | sort
   true
}

echo_cmd()
{
   echo "$@"
   "$@" || exit
}

echo_cat()
{
   echo cat "$1"\>\>$2
   cat $1>>$2 || exit
}

echo_echo1()
{
   echo echo "$1"\>\>$2
   echo $1>$2 || exit
}

echo_echo()
{
   echo echo "$1"\>\>$2
   echo $1>>$2 || exit
}

echo_set()
{
   echo $1="$2"
   eval $1="\$2"
}

touch_if()
{
   if ! [ -f $1 ]; then
      echo_cmd mkdir -p $(dirname $1)
      echo_cmd touch $1
   fi
}

copy_if()
{
   local src="$1"
   local dest="$2"

   mkdir -p $(dirname $dest)
   if [ -f $src ]; then
      echo_cmd cp $src $dest
   else
      touch_if $dest
   fi
}

copy_to_if()
{
   local src="$1"
   local dest="$2"
   local end_name="$(basename $src)"

   copy_if $src $dest/$end_name
}

move_to()
{
   # dest

   if [ "$1" ] && [ -e "$1" ]; then
      echo_cmd mv "$@" $dest/
   fi
}

make_sh_link()
{
   local source="$1"
   local target="$2"

   rm -r -f $source
   if [ -f $target ]; then
      echo \#!/bin/sh>$source
      echo exec $target \"\$@\">>$source
      chmod a+x $source
   fi
}

is_not_script()
{
   local a="$1"

   [ -f $a ] && [ "$(head -1 $a)" != "#!/bin/sh" ] && true
   false
}

fix_platform_folder()
{
   # platform
   # tuple
   local a
   local b
   local c
   local d
   local e
   local qu=\"
   local gt=\>
   local s=\$

   for a in $(ls src/stop/* 2>/dev/null); do
      b=${a#src/stop/}
      c="$platform/bin/$tuple-$b-"
      d="$platform/$tuple/bin/$b"
      if [ -f $d ]; then
         if [ -f $c ]; then
            echo_cmd rm $d
         else
            echo_cmd mv $d $c
         fi
      fi
      d="$platform/bin/$tuple-$tuple-$b-"
      [ -f $d ] && echo_cmd rm $d
      d="build/bin/$tuple-$b"
      is_not_script $d && echo_cmd rm $d
      if [ -f $c ] && [ ! -f $d ]; then
         echo_echo "#!/bin/sh" $d
         echo_echo "ptlinsdk=$qu$ptlinsdk$qu" $d
         echo_echo "tuple=$qu$tuple$qu" $d
         echo_echo "platform=$qu$platform$qu" $d
         echo_echo "target_cflags=$qu$target_cflags$qu" $d
         echo_echo "target_ldflags=$qu$target_ldflags$qu" $d
         echo_echo "sys_path=$qu$sys_path$qu" $d
         echo_echo "export PATH=$qu$ptlinsdk/build/bin:$ptlinsdk/build/bin/$platform:$ptlinsdk/build/bin/stop:$sys_path$qu" $d
         echo_echo "export LD_LIBRARY_PATH=$qu$ptlinsdk/build/usr/lib$qu" $d
         echo_echo "c=$qu$ptlinsdk/$c$qu" $d
         [ -f src/target/$b ] && echo_cat "src/target/$b" $d
         echo_echo "[ ${s}ptlinsdk_log ] && echo ${s}c $qu$s@$qu$gt$gt${s}ptlinsdk_log" $d
         echo_echo "exec ${s}c $qu$s@$qu" $d
         echo_cmd chmod a+x $d
      fi
      e="build/bin/$platform/$b"
      is_not_script $e && echo_cmd rm $e
      [ -f $e ] || [ ! -f $d ] || echo_cmd cp $d $e
      # e="$platform/bin/$b"
      # is_not_script $e && echo_cmd rm $e
      # [ -f $e ] || [ $platform = build ] || [ ! -f $d ] || echo_cmd cp $d $e
   done
   c="$ptlinsdk/build/bin/$tuple-gcc"
   make_sh_link build/bin/$tuple-cc $c
   make_sh_link build/bin/$tuple/cc $c
   c="$ptlinsdk/build/bin/$tuple-g++"
   make_sh_link build/bin/$tuple-c++ $c
   make_sh_link build/bin/$tuple/c++ $c
   c="$platform/lib/libiberty.a"
   [ -f $c ] && echo_cmd rm $c
}

setup_platform_folder()
{
   # platform
   # tuple
   # build_mac_arch
   local a
   local b
   local c
   local d
   local qu=\"
   local gt=\>
   local s=\$

   if ! [ -d $ptlinsdk/obj/$tuple ]; then
      echo_cmd mkdir -p $ptlinsdk/$platform/usr
      echo_cmd cd $ptlinsdk/build
      test $platform = build || echo_cmd ln -s ../$platform $tuple
      echo_cmd cd $ptlinsdk/$platform/usr
      echo_cmd mkdir -p libexec
      echo_cmd mkdir -p man
      echo_cmd mkdir -p info
      echo_cmd mkdir -p include
      echo_cmd mkdir -p bin
      echo_cmd mkdir -p lib/pkgconfig
      echo_cmd ln -s lib lib64
      echo_cmd mkdir -p share
      echo_cmd cd share
      echo_cmd ln -s ../man
      echo_cmd ln -s ../info
      echo_cmd ln -s ../lib/pkgconfig
      echo_cmd cd $ptlinsdk/$platform/usr
      echo_cmd mkdir -p $tuple
      echo_cmd cd $tuple
      echo_cmd ln -s ../include
      echo_cmd ln -s ../lib
      echo_cmd ln -s ../lib lib64
      echo_cmd ln -s ../libexec
      echo_cmd ln -s ../share
      echo_cmd ln -s ../man
      echo_cmd ln -s ../info
      echo_cmd cd $ptlinsdk/$platform
      echo_cmd ln -s usr/libexec
      echo_cmd ln -s usr/share
      echo_cmd ln -s usr/man
      echo_cmd ln -s usr/info
      echo_cmd ln -s usr/include
      echo_cmd ln -s usr/bin
      echo_cmd ln -s usr/lib
      echo_cmd ln -s usr/lib lib64
      echo_cmd mkdir -p $tuple
      echo_cmd cd $tuple
      echo_cmd ln -s ../usr/include
      echo_cmd ln -s ../usr/lib
      echo_cmd ln -s ../usr/lib lib64
      echo_cmd ln -s ../usr/libexec
      echo_cmd ln -s ../usr/share
      echo_cmd ln -s ../usr/man
      echo_cmd ln -s ../usr/info
      echo_cmd cd $ptlinsdk/build/bin
      echo_cmd mkdir -p $platform
      echo_cmd ln -s $platform $tuple
      echo_cmd cd $ptlinsdk
      if [ $platform = build ]; then
         echo_cmd mkdir -p build/bin/stop
         for a in $(ls src/stop/* 2>/dev/null); do
            b="${a#src/stop/}"
            echo_cmd cp $a build/bin/stop/$b
         done
      fi
      echo_cmd mkdir -p obj/$tuple
   fi
   if [ $platform = build ]; then
      for a in $(ls src/$build_sys/* 2>/dev/null); do
         b="${a#src/$build_sys/}"
         d="build/bin/$tuple-$b"
         is_not_script $d && echo_cmd rm $d
         if [ ! -f $d ]; then
            echo_echo "#!/bin/sh" $d
            echo_echo "ptlinsdk=$qu$ptlinsdk$qu" $d
            echo_echo "tuple=$qu$tuple$qu" $d
            echo_echo "build_arch=$qu$build_mac_arch$qu" $d
            echo_echo "build_cflags=$qu$build_cflags$qu" $d
            echo_echo "build_ldflags=$qu$build_ldflags$qu" $d
            echo_echo "sys_path=$qu$sys_path$qu" $d
            echo_echo "export PATH=$qu$ptlinsdk/build/bin:$ptlinsdk/build/bin/build:$ptlinsdk/build/bin/stop:$sys_path$qu" $d
            echo_echo "export LD_LIBRARY_PATH=$qu$ptlinsdk/build/usr/lib$qu" $d
            echo_echo "c=$qu$ptlinsdk/$d-$qu" $d
            echo_cat $a $d
            echo_echo "[ ${s}ptlinsdk_log ] && echo ${s}c $qu$s@$qu$gt$gt${s}ptlinsdk_log" $d
            echo_echo "exec ${s}c $qu$s@$qu" $d
            echo_cmd chmod a+x $d
            echo_cmd cp $d "build/bin/$tuple/$b"
         fi
      done
   fi
   fix_platform_folder
}

copy_current_folder()
{
   local dest="$1"

   echo find . \| cpio -p $dest
   find . | cpio -p -u $dest || exit
}

cp_core_system_libs()
{
   local src="$1"
   local name

   if cd $src; then
      for name in "libc-*" "libc.*" "libm-*" "libm.*" "librt*" "libdl*" "libpthread*" "libstdc*" "*.o" "libexpat*" "libbind*" "libresolv*" ; do
         find . -maxdepth 1 -name "$name" | cpio -pv $ptlinsdk/build/lib
      done
   fi
   cd $ptlinsdk || exit
}

setup_sys_libs()
{
   # Debian multiarch breaks building/using alternate
   # compiler. Redhat use of lib64 also creates issues.
   local done="obj/$tuple/fix-multiarch-done"
   if [ ! -f $done  ]; then
      echo
      echo --xbt-- flattening multiarch system libs/includes
      echo
      if [ -d /lib/$gnu_arch-linux-gnu ]; then
         echo_cmd cp_core_system_libs /usr/lib/$gnu_arch-linux-gnu
         echo_cmd cp_core_system_libs /lib/$gnu_arch-linux-gnu
         if cd /usr/include/$gnu_arch-linux-gnu; then
               # need xlocal.h in folder with "sys" for X11/Xlocale.h
            echo_cmd cp /usr/include/xlocale.h $ptlinsdk/build/include
            echo_cmd copy_current_folder $ptlinsdk/build/include
         fi
      fi
      echo_cmd cp_core_system_libs /lib
      echo_cmd cp_core_system_libs /lib64
      echo_cmd cd $ptlinsdk
      mkdir -p obj/$tuple
      echo_cmd touch $done
   fi
}

tar_download()
{
   local name="$1"
   local ver
   local src_name
   local package
   local url
   local suffix

   eval ver=\"\$\{${name}_ver\}\"
   eval src_name=\"\$\{${name}_name\}\"
   if ! [ $src_name ]; then
      src_name="${name}-$ver"
      eval ${name}_name=\"$src_name\"
   fi
   eval package=\"\$\{${name}_package\}\"
   if ! [ $package ]; then
      eval suffix=\"\$\{${name}_suffix\}\"
      if ! test $suffix; then
         suffix=".tar.gz"
      fi
      package="$src_name$suffix"
      eval ${name}_package=\"$package\"
   fi
   eval url=\"\$\{${name}_url\}\"
   if ! [ $url ]; then
      eval url=\"\$\{${name}_url_\}\"
      [ $url ] || url="$gnumirror_url/$name/"
      url="$url$package"
      eval ${name}_url=\"$url\"
   fi
   if  ! [ -f tars/$package ]; then
      echo
      echo --xbt-- downloading $package
      echo
      if [ ! $(PATH="$ptlinsdk/build/bin:$PATH" which wget) ]; then
         echo error: wget not available, unable to download package
         echo to proceed, download $url to $ptlinsdk/tars/$package
         exit
      fi
      PATH="$ptlinsdk/build/bin:$PATH" echo_cmd wget --no-check-certificate -O tars/${package}_part "$url"
      mv tars/${package}_part tars/$package
   fi
}

tar_extract()
{
   # name
   eval local src_name=\"\$\{${name}_name\}\"
   eval local package=\"\$\{${name}_package\}\"
   local content_name

   [ $name ] || echo_cmd false name not defined in tar_extract
   [ $src_name ] || echo_cmd false ${name}_name not defined in tar_extract
   [ $package ] || echo_cmd false ${name}_package not defined in tar_extract
   if ! [ -d obj/$src_name ]; then
      echo
      echo --xbt-- extracting $package
      echo
      echo_cmd rm -r -f obj/$src_name
      echo_cmd cd obj/part
      echo_cmd rm -r -f $src_name*
      case $package in
      (*.tar.gz|*.tgz)
         echo_cmd tar xzf $ptlinsdk/tars/$package
         ;;
      (*.tar.bz2)
         echo_cmd tar xjf $ptlinsdk/tars/$package
         ;;
      (*)
         echo_cmd false do not know how to extract $package
      esac
      content_name="$src_name"
      [ -d $content_name ] || unset content_name
      [ $content_name ] || content_name="$(ls -d $src_name* 2>/dev/null)"
      [ -d $content_name ] || unset content_name
      [ $content_name ] || content_name="$(ls -d $name* 2>/dev/null)"
      [ -d $content_name ] || unset content_name
      [ $content_name ] || echo_cmd false $package content folder is missing
      if [ -f $ptlinsdk/src/patch/$src_name.patch ]; then
         echo_cmd cd $content_name
         echo_cmd patch -t -p1 -i $ptlinsdk/src/patch/$src_name.patch
      fi
      echo_cmd cd $ptlinsdk
      echo_cmd mv obj/part/$content_name obj/$src_name
   fi
}

config_gcc()
{
   unset gcc_name gcc_package gcc_url
   gcc_ver=$1
   eval gcc_ver=\"\$\{gcc_${1}_ver\}\"
   gcc_url_="$gnumirror_url/gcc/gcc-$gcc_ver/"
   tar_download gcc
}

autotool_configure()
{
   # platform
   # tuple
   # name
   # environment
   local src_name
   local src_folder
   local obj_folder

   eval src_name=\"\$\{${name}_name\}\"
   src_folder="obj/$src_name"
   obj_folder="obj/$tuple/$obj_name"
   tar_extract
   if ! [ -f $obj_folder/Makefile ]; then
      echo
      echo --xbt-- configuring $obj_name for $tuple
      echo
      echo_cmd rm -r -f $obj_folder
      echo_cmd mkdir -p $obj_folder
      if [ -f src/config/$name ]; then
         echo_cat src/config/$name $obj_folder/config.cache
      fi
      if [ -d src/config/$name ]; then
         echo_cmd cd src/config/$name
         copy_current_folder $ptlinsdk/$obj_folder
         echo_cmd cd $ptlinsdk
      fi
      echo_cmd cd $obj_folder
      echo_env
      echo_cmd $ptlinsdk/obj/$src_name/configure --cache-file=config.cache "$@"
      echo_cmd cd $ptlinsdk
   fi
}

autotool_make()
{
   # environment
   # platform
   # tuple
   # name
   # obj_name
   # make_goal
   # install_goal
   local done

   done="obj/$tuple/$obj_name/make-done"
   if ! [ -f $done  ]; then
      echo_cmd autotool_configure "$@"
      echo_cmd cd obj/$tuple/$obj_name
      if [ $make_goal ]; then
         echo
         echo --xbt-- making $obj_name for $tuple
         echo
         echo_env
         echo_cmd make -j $CONCURRENCY $make_goal
      fi
      if [ $install_goal ]; then
         echo
         echo --xbt-- installing $obj_name for $tuple
         echo
         echo_env
         echo_cmd make $install_goal
      fi
      cd $ptlinsdk
      fix_platform_folder
      echo_cmd touch $done
   fi
}

autotool_make_build()
{
   # platform=build
   # tuple=$build_tuple
   # name
   eval local obj_name=\"\$\{${name}_name\}\"
   local make_goal="all"
   local install_goal="install"

   setup_env $tuple
   autotool_make --build=$build_tuple --host=$build_tuple --prefix=$ptlinsdk/build "$@"
}

autotool_make_target()
{
   # platform
   # tuple
   # name
   eval local src_name=\"\$\{${name}_name\}\"
   local obj_name="$src_name"
   local make_goal="all"
   local install_goal="install"

   setup_env $tuple
   autotool_make --build=$build_tuple --host=$tuple --prefix=$ptlinsdk/$platform "$@"
}

autotool_make_shared()
{
   setup_env $tuple
   autotool_make_target --enable-shared --with-shared --disable-static --without-static "$@"
}

autotool_make_static()
{
   setup_env $tuple
   autotool_make_target --disable-shared --without-shared --enable-static --with-static "$@"
}

configure_make()
{
   # environment
   # platform
   # tuple
   # name
   # src_name
   # obj_name
   eval local src_name=\"\$\{${name}_name\}\"
   local obj_name="$src_name"
   local src_folder="obj/$src_name"
   local obj_folder="obj/$tuple/$obj_name"
   local done

   tar_extract
   done="$obj_folder/configure-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- configuring $obj_name for $tuple
      echo
      echo_cmd rm -r -f $obj_folder
      echo_cmd mkdir -p $obj_folder
      echo_cmd cd $src_folder
      echo_cmd copy_current_folder $ptlinsdk/$obj_folder
      echo_cmd cd $ptlinsdk
      echo_cmd cd $obj_folder
      if [ -f configure ]; then
         echo_env
         echo_cmd ./configure --prefix=$ptlinsdk/$platform "$@"
      fi
      echo_cmd cd $ptlinsdk
      echo_cmd touch $done
   fi
   done="obj/$tuple/$obj_name/make-done"
   if ! [ -f $done ]; then
      echo_cmd cd obj/$tuple/$obj_name
      echo
      echo --xbt-- making $obj_name for $tuple
      echo
      echo_env
      echo_cmd make -j $CONCURRENCY
      echo
      echo --xbt-- installing $obj_name for $tuple
      echo
      echo_env
      [ -f configure ] || echo_cmd make PREFIX=$ptlinsdk/$platform install
      [ -f configure ] && echo_cmd make install
      cd $ptlinsdk
      fix_platform_folder
      echo_cmd touch $done
   fi
}

configure_make_build()
{
   # platform=build
   # tuple=$build_tuple
   # name

   setup_env $build_tuple
   configure_make "$@"
}

configure_make_target()
{
   # platform=
   # tuple=
   # name

   setup_env $tuple
   configure_make "$@"
}

make_linux_headers()
{
   # platform
   # tuple
   # lin_arch
   local name="linux"
   local src_name="$linux_name"
   local src_folder="obj/$src_name"
   # local obj_folder="$ptlinsdk/obj/$tuple/$src_name"
   local done

   done="obj/$tuple/linux-headers-done"
   if ! [ -f $done ]; then
      tar_extract
      echo
      echo --xbt-- installing $src_name headers for $tuple
      echo
      # echo_cmd rm -r -f $obj_folder
      # echo_cmd mkdir -p $obj_folder
      setup_env $build_tuple
      echo_cmd cd $src_folder
      # echo_cmd make O=$obj_folder ARCH=$lin_arch CROSS_COMPILE=$tuple- INSTALL_HDR_PATH=$ptlinsdk/$platform/usr headers_install
      echo_cmd make ARCH=$lin_arch CROSS_COMPILE=$tuple- INSTALL_HDR_PATH=$ptlinsdk/$platform/usr headers_install
      echo_cmd cd $ptlinsdk
      # This header expected by glibc 2.6 for arm, but absent in newer kernels.
      echo_cmd touch $platform/usr/include/asm/page.h
      echo_cmd touch $done
   fi
}

make_rhel()
{
   # name
   # config
   # arch
   # gnu_arch
   # lin_arch
   # tuple
   local platform="lin-$arch"
   eval local src_name=\"\$\{${name}_name\}\"
   eval local package=\"\$\{${name}_package\}\"
   local src_folder_="build/usr/src"
   local src_folder="$src_folder_/$src_name"
   local build_folder
   local strip_folder
   local done
   local host_libs=""
   local dot_config=""
   local dash_config=""
   local build_name
   local config_file
   eval local rhel_ver=\"\$\{${name}_ver\}\"

   if [ "$config" ]; then
      dot_config=".$config"
      dash_config="-$config"
   fi
   # /lib/modules/2.6.32-x.el6.debug.i686/build
   build_name="$rhel_ver.$gnu_arch$dot_config"
   # kernel-2.6.18-i686-PAE.config
   config_file="kernel-*-$gnu_arch$dash_config.config"
   if test "$build_sys" = mac; then
      host_libs="-lintl"
   fi
   build_folder="$ptlinsdk/build/lib/modules/$build_name/build"
   strip_folder="obj/lib/modules/$build_name/build"

   if ! [ -d $src_folder ]; then
      echo
      echo --xbt-- extracting $src_name
      echo
      echo_cmd mkdir -p $src_folder_/part
      echo_cmd cd $src_folder_/part
      echo_cmd tar xzf $ptlinsdk/tars/$package
      if [ -f $ptlinsdk/src/patch/$src_name.patch ]; then
         echo_cmd cd $src_name
         echo_cmd patch -t -p1 -i $ptlinsdk/src/patch/$src_name.patch
      fi
      echo_cmd cd $ptlinsdk
      echo_cmd mv $src_folder_/part/$src_name $src_folder_/
      rmdir $src_folder_/part 2>/dev/null
   fi
   done="$build_folder/configure-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- configuring linux-$build_name
      echo
      echo_cmd rm -r -f $build_folder
      echo_cmd mkdir -p $build_folder
      echo_cmd cp $src_folder/configs/$config_file $build_folder/.config
      setup_env $build_tuple
      echo_cmd make HOST_LOADLIBES="$host_libs" -C $src_folder ARCH=$lin_arch CROSS_COMPILE=$tuple- O=$build_folder oldconfig
      echo_cmd touch $done
   fi
   done="$build_folder/make-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- making linux-$build_name
      echo
      setup_env $build_tuple
      echo_cmd make HOST_LOADLIBES="$host_libs" -j $CONCURRENCY -C $src_folder ARCH=$lin_arch CROSS_COMPILE=$tuple- O=$build_folder vmlinux
      echo_cmd touch $done
   fi
   done="$build_folder/strip-done"
   if ! [ -f $done ]; then
      echo_cmd mkdir -p $strip_folder
      dest=$strip_folder move_to $build_folder/..tmp*
      dest=$strip_folder move_to $build_folder/.config.old
      dest=$strip_folder move_to $build_folder/.tmp*
      dest=$strip_folder move_to $build_folder/.vmlinux.cmd
      dest=$strip_folder move_to $build_folder/block
      dest=$strip_folder move_to $build_folder/crypto
      dest=$strip_folder move_to $build_folder/drivers
      dest=$strip_folder move_to $build_folder/fs
      dest=$strip_folder move_to $build_folder/init
      dest=$strip_folder move_to $build_folder/ipc
      dest=$strip_folder move_to $build_folder/kernel
      dest=$strip_folder move_to $build_folder/lib
      dest=$strip_folder move_to $build_folder/mm
      dest=$strip_folder move_to $build_folder/net
      dest=$strip_folder move_to $build_folder/samples
      dest=$strip_folder move_to $build_folder/security
      dest=$strip_folder move_to $build_folder/sound
      dest=$strip_folder move_to $build_folder/usr
      dest=$strip_folder move_to $build_folder/vmlinux.*
      echo_cmd touch $done
   fi
}

strip_rhel()
{
   # name
   eval local src_name=\"\$\{${name}_name\}\"
   local src_folder="build/usr/src/$src_name"
   local strip_folder="obj/usr/src/$src_name"
   local done="$src_folder/strip-done"

   if ! [ -f $done ]; then
      echo_cmd mkdir -p $strip_folder
      dest=$strip_folder move_to $src_folder/.gitignore
      dest=$strip_folder move_to $src_folder/COPYING
      dest=$strip_folder move_to $src_folder/CREDITS
      dest=$strip_folder move_to $src_folder/Documentation
      dest=$strip_folder move_to $src_folder/MAINTAINERS
      dest=$strip_folder move_to $src_folder/README
      dest=$strip_folder move_to $src_folder/REPORTING-BUGS
      dest=$strip_folder move_to $src_folder/block
      dest=$strip_folder move_to $src_folder/configs
      dest=$strip_folder move_to $src_folder/crypto
      dest=$strip_folder move_to $src_folder/drivers
      dest=$strip_folder move_to $src_folder/*.sh
      dest=$strip_folder move_to $src_folder/fs
      dest=$strip_folder move_to $src_folder/init
      dest=$strip_folder move_to $src_folder/ipc
      dest=$strip_folder move_to $src_folder/kernel
      dest=$strip_folder move_to $src_folder/lib
      dest=$strip_folder move_to $src_folder/mm
      dest=$strip_folder move_to $src_folder/net
      dest=$strip_folder move_to $src_folder/*.txt
      dest=$strip_folder move_to $src_folder/samples
      dest=$strip_folder move_to $src_folder/security
      dest=$strip_folder move_to $src_folder/sound
      dest=$strip_folder move_to $src_folder/tools
      dest=$strip_folder move_to $src_folder/usr
      dest=$strip_folder move_to $src_folder/.mailmap
      dest=$strip_folder move_to $src_folder/check-kabi
      dest=$strip_folder move_to $src_folder/config-*
      dest=$strip_folder move_to $src_folder/extract.pub
      dest=$strip_folder move_to $src_folder/extrakeys.pub
      dest=$strip_folder move_to $src_folder/firmware
      dest=$strip_folder move_to $src_folder/kabitool
      dest=$strip_folder move_to $src_folder/merge.pl
      dest=$strip_folder move_to $src_folder/perf
      dest=$strip_folder move_to $src_folder/pubring.gpg
      dest=$strip_folder move_to $src_folder/random_seed
      dest=$strip_folder move_to $src_folder/temp*
      dest=$strip_folder move_to $src_folder/virt
      echo_cmd touch $done
   fi
}

make_gcc()
{
   # platform
   # tuple
   # obj_name
   # rpath
   local name="gcc"
   local make_goal="all"
   local install_goal="install"

   setup_env
   export LD_LIBRARY_PATH="$rpath"
   export CPPFLAGS="$CPPFLAGS_FOR_BUILD"
   export CFLAGS="$CFLAGS_FOR_BUILD"
   export CXXFLAGS="$CXXFLAGS_FOR_BUILD"
   export LIBCFLAGS="$LIBCFLAGS_FOR_BUILD"
   export LIBCXXFLAGS="$CXXFLAGS_FOR_BUILD"
   autotool_make cross_compiling=yes --host=$build_tuple --build=$build_tuple --target=$tuple --prefix=$ptlinsdk/$platform --includedir=$ptlinsdk/$platform/usr/include --program-prefix=$tuple- --program-suffix=- --with-mpfr=$ptlinsdk/build --with-gmp=$ptlinsdk/build --disable-nls --disable-multilib --disable-lto --disable-libssp --disable-libmudflap --disable-libquadmath "$@"
}

glibc_make_prep()
{
   # platform
   local libs="libc.so libBrokenLocale.a libanl.a libbsd-compat.a libc.a libc_nonshared.a libcrypt.a libdl.a libg.a libieee.a libm.a libmcheck.a libpthread.a libpthread_nonshared.a libresolv.a librt.a"
   local a

   # Delete any already installed libs, dummy libs, headers, from
   # previous pass install (or failed install). Otherwise may get
   # found during due to the tool wrappers inserting paths outside
   # of the glibc makefiles control.
   for a in $libs; do
      a="$platform/lib/$a"
      [ ! -f $a ] || echo_cmd rm $a
   done
   rm -f "$platform/include/gnu/stubs.h"
   touch "$platform/include/gnu/stubs.h"
}

make_glibc_headers()
{
   # platform
   # tuple
   local name="glibc"
   local src_name="$glibc_name"
   local obj_name="${glibc_name}_pass1"
   local src_folder="obj/$src_name"
   local obj_folder="obj/$tuple/$obj_name"
   local done

   if ! [ -d obj/$glibc_name ]; then
      [ $glibc_name ] || echo_cmd false glibc_name not defined in make_glibc_headers
      echo_cmd rm -r -f obj/$glibc_name
      echo_cmd cd obj/part
      echo_cmd rm -r -f $glibc_name
      echo_cmd tar xzf $ptlinsdk/tars/$glibc_package
      echo_cmd rm -r -f $glibc_ports_name
      echo_cmd tar xzf $ptlinsdk/tars/$glibc_ports_package
      echo_cmd mv $glibc_ports_name $glibc_name/ports
      if [ -f $ptlinsdk/src/patch/$glibc_name.patch ]; then
         echo_cmd cd $glibc_name
         echo_cmd patch -t -p1 -i $ptlinsdk/src/patch/$glibc_name.patch
         echo_cmd cd ..
      fi
      echo_cmd cd $ptlinsdk
      echo_cmd mv obj/part/$glibc_name obj
   fi
   # Build environment used instead of target environment
   # since target tools aren't available yet and just need
   # the libc headers. Will do it right later.
   setup_env $build_tuple
   done="$obj_folder/configure-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- configuring $obj_name for $tuple
      echo
      rm -r -f $obj_folder
      echo_cmd mkdir -p $obj_folder
      echo_cat src/config/glibc_headers $obj_folder/config.cache
      echo_env
      glibc_make_prep
      touch_if $ptlinsdk/$platform/usr/include/limits.h
      echo_cmd cd $obj_folder
      echo_cmd $ptlinsdk/obj/$src_name/configure --cache-file=config.cache --build=$build_tuple --host=$tuple --prefix=/usr --with-headers=$ptlinsdk/$platform/usr/include --disable-profile --disable-nls --disable-multilib --disable-static --without-gd --without-cvs --enable-addons=nptl,ports $glibc_configure
      echo_cmd cd $ptlinsdk
      echo_cmd touch $done
   fi
   done="$obj_folder/make-done"
   if [ ! -f $done ]; then
      echo
      echo --xbt-- installing $obj_name for $tuple
      echo
      echo_env
      glibc_make_prep
      echo_cmd cd $obj_folder
      echo_cmd make install_root=$ptlinsdk/$platform install-headers
      echo_cmd cd $ptlinsdk
      copy_if $obj_folder/bits/stdio_lim.h $platform/usr/include/bits/stdio_lim.h
      copy_if $src_folder/include/features.h $platform/usr/include/features.h
      touch_if $platform/usr/include/gnu/stubs.h
      echo_cmd touch $done
   fi
}

make_glibc_startup()
{
   # platform
   # tuple
   local name="glibc"
   local src_name="$glibc_name"
   local obj_name="${glibc_name}_pass2"
   local src_folder="obj/$src_name"
   local obj_folder="obj/$tuple/$obj_name"
   local done

   setup_env $tuple
   done="$obj_folder/configure-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- configuring $obj_name for $tuple
      echo
      glibc_make_prep
      touch_if $platform/usr/include/sys/cdefs.h
      touch_if $platform/usr/include/bits/posix1_lim.h
      touch_if $platform/usr/include/bits/posix2_lim.h
      rm -r -f $obj_folder
      echo_cmd mkdir -p $obj_folder
      if [ -f src/config/glibc ]; then
         echo_cat src/config/glibc $obj_folder/config.cache
      fi
      echo_cmd cd $obj_folder
      echo_env
      echo_cmd $ptlinsdk/obj/$src_name/configure --cache-file=config.cache --build=$build_tuple --host=$tuple --prefix=/usr --with-headers=$ptlinsdk/$platform/usr/include --disable-profile --disable-nls --disable-multilib --disable-static --without-gd --without-cvs --enable-addons=nptl,ports $glibc_configure
      cd $ptlinsdk
      echo_cmd touch $done
   fi
   done="$obj_folder/make-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- making $obj_name for $tuple
      echo
      glibc_make_prep
      echo_cmd cd $obj_folder
      echo_cmd make -j $CONCURRENCY csu/subdir_lib
      echo_cmd cd $ptlinsdk
      echo
      echo --xbt-- installing $obj_name for $tuple
      echo
      # temporary dummy libc.
      echo_cmd build/bin/$tuple-gcc -nostdlib -nostartfiles -shared -x c /dev/null -o $platform/lib/libc.so
      for a in $(ls $obj_folder/csu/crt*.o 2>/dev/null); do
         echo_cmd cp $a $platform/lib
      done
      echo_cmd touch $done
   fi
}

make_glibc_final()
{
   # platform
   # tuple
   local name="glibc"
   local src_name="$glibc_name"
   local obj_name="${glibc_name}_final"
   local src_folder="obj/$src_name"
   local obj_folder="obj/$tuple/$obj_name"
   local done
   local c

   setup_env $tuple
   done="$obj_folder/configure-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- configuring $obj_name for $tuple
      echo
      rm -r -f $obj_folder
      echo_cmd mkdir -p $obj_folder
      if [ -f src/config/glibc ]; then
         echo_cat src/config/glibc $obj_folder/config.cache
      fi
      glibc_make_prep
      echo_cmd cd $obj_folder
      echo_env
      echo_cmd $ptlinsdk/obj/$src_name/configure --cache-file=config.cache --build=$build_tuple --host=$tuple --prefix=/usr --with-headers=$ptlinsdk/$platform/usr/include --disable-profile --disable-nls --disable-multilib --disable-static --without-gd --without-cvs --enable-addons=nptl,ports $glibc_configure
      cd $ptlinsdk
      echo_cmd touch $done
   fi
   done="$obj_folder/make-done"
   if ! [ -f $done ]; then
      echo
      echo --xbt-- making $obj_name for $tuple
      echo
      glibc_make_prep
      echo_cmd cd $obj_folder
      echo_cmd make -j $CONCURRENCY
      echo
      echo --xbt-- installing $obj_name for $tuple
      echo
      echo_cmd make install_root=$ptlinsdk/$platform install
      echo_cmd cd $ptlinsdk
      echo_cmd touch $done
   fi
}

make_gcc_toolchain()
{
   # arch
   # gnu_arch
   # lin_arch
   # tuple
   local platform="lin-$arch"
   local name
   local obj_name
   local make_goal
   local install_goal
   local b
   local c
   local done

   setup_platform_folder

   setup_env $build_tuple
   name="binutils"
   obj_name="$binutils_name"
   make_goal="all"
   install_goal="install"
   autotool_make --build=$build_tuple --host=$build_tuple --target=$tuple --prefix=$ptlinsdk/$platform --program-prefix=$tuple- --program-suffix=- --with-lib-path=$ptlinsdk/$platform/usr/lib --enable-64-bit-bfd --with-sysroot --disable-multilib --disable-nls --disable-lto

   make_linux_headers

   make_glibc_headers

   config_gcc pass1
   obj_name="${gcc_name}_pass1"
   make_gcc --disable-shared --disable-libgomp --enable-languages=c --disable-decimal-float $gcc_configure
   for a in $(ls $platform/lib/gcc/$tuple/$gcc_ver/include-fixed/*.h 2>/dev/null); do
      b="${a#$platform/lib/gcc/$tuple/$gcc_ver/include-fixed/}"
      c="$platform/lib/gcc/$tuple/$gcc_ver/include/$b"
      copy_if $a $c
   done

   make_glibc_startup

   config_gcc pass2
   obj_name="${gcc_name}_pass2"
   make_gcc --disable-libgomp --enable-languages=c $gcc_configure

   make_glibc_final

   config_gcc final
   obj_name="${gcc_name}_final"
   make_gcc --enable-__cxa_atexit --enable-languages=c,c++ $gcc_configure

   setup_env $build_tuple
   name=libtool
   obj_name="$libtool_name"
   make_goal="all"
   install_goal="install"
   autotool_make --build=$build_tuple --host=$build_tuple --target=$tuple --prefix=$ptlinsdk/$platform --program-prefix=$tuple- --program-suffix=-

   setup_env $build_tuple
   name="pkg_config"
   obj_name="$pkg_config_name"
   make_goal="all"
   install_goal="install"
   autotool_make --build=$build_tuple --host=$build_tuple --target=$tuple --prefix=$ptlinsdk/$platform --program-prefix=$tuple- --program-suffix=-

   done="obj/mac-linux-compat-includes-done"
   if [ ! -f $done ] && [ $build_sys = mac ] && [ $build_arch = $arch ]; then
      # On mac build systems, add some linux system includes
      # to the build include folder to allow cross-compiling
      # linux kernel.
      echo_cmd mkdir -p build/usr/include/sys
      echo_cmd mkdir -p build/usr/include/asm
      echo_cmd mkdir -p build/usr/include/asm-generic
      echo_cmd mkdir -p build/usr/include/asm-$lin_arch
      echo_cmd mkdir -p build/usr/include/dispatch
      touch_if build/usr/include/features.h
      copy_to_if src/malloc.h build/usr/include
      copy_to_if src/byteswap.h build/usr/include
      copy_to_if src/endian.h build/usr/include
      copy_if    src/linux_types.h build/usr/include/asm/types.h
      copy_if    src/linux_types.h build/usr/include/asm-generic/types.h
      copy_if    src/linux_types.h build/usr/include/asm-$lin_arch/types.h
      copy_if    src/linux_types.h build/usr/include/linux/types.h
      copy_to_if $platform/usr/include/elf.h build/usr/include
      copy_to_if $platform/usr/include/sys/sysmacros.h build/usr/include/sys
      copy_to_if $platform/usr/include/asm/swab.h build/usr/include/asm
      # copy_to_if $platform/usr/include/asm/types.h build/usr/include/asm
      copy_to_if $platform/usr/include/asm/byteorder.h build/usr/include/asm
      # copy_to_if $platform/usr/include/asm-generic/types.h build/usr/include/asm-generic
      # copy_to_if $platform/usr/include/asm-$lin_arch/types.h build/usr/include/asm-$lin_arch
      copy_to_if $platform/usr/include/asm-$lin_arch/byteorder.h build/usr/include/asm-$lin_arch
      copy_to_if $platform/usr/include/linux/stddef.h build/usr/include/linux
      copy_to_if $platform/usr/include/linux/swab.h build/usr/include/linux
      copy_to_if $platform/usr/include/linux/byteorder/little_endian.h build/usr/include/linux/byteorder
      copy_to_if $platform/usr/include/linux/byteorder/big_endian.h build/usr/include/linux/byteorder
      copy_to_if $platform/usr/include/linux/byteorder/swab.h build/usr/include/linux/byteorder
      copy_to_if $platform/usr/include/linux/byteorder/generic.h build/usr/include/linux/byteorder
      echo_cmd touch $done
   fi
}

make_build_includes()
{
   done="obj/mac-gcc-compat-includes-done"
   if [ ! -f $done ] && [ $build_sys = mac ]; then
      # On mac build systems, add some alternate system headers
      # to deal with clang/gcc compat issues.
      echo_cmd mkdir -p build/usr/include/dispatch
      copy_if    src/dispatch_object.h build/usr/include/dispatch/object.h
      copy_to_if src/Availability.h build/usr/include
      echo_cmd touch $done
   fi
}

make_x11()
{
   name=xcb_proto autotool_make_static
   name=xproto autotool_make_static
   name=xextproto autotool_make_static
   name=kbproto autotool_make_static
   name=inputproto autotool_make_static
   name=renderproto autotool_make_static
   name=libpthread_stubs autotool_make_static
   name=libxau autotool_make_shared
   name=libxcb autotool_make_shared
   name=xtrans autotool_make_static
      # Work around for glibc/X11 header name conflict in libx11
      # needs a dummy xlocale.h file if not provided by libc.
   if test $platform = build; then
      test -e /usr/include/xlocale.h || touch $platform/include/xlocale.h
   else
      test -e $platform/include/xlocale.h || touch $platform/include/xlocale.h
   fi
   name=libx11 autotool_make_shared --disable-malloc0returnsnull
   name=libxext autotool_make_shared --disable-malloc0returnsnull
   name=libxrender autotool_make_shared --disable-malloc0returnsnull
}

make_gtk()
{
   name=libffi autotool_make_shared
   if test $platform = build; then
      name=libiconv autotool_make_shared
      name=glib autotool_make_shared --with-libiconv=gnu
   else
      name=glib autotool_make_shared --with-libiconv=no
   fi
   name=atk autotool_make_shared
   name=libpng autotool_make_shared
   name=freetype CONCURRENCY=1 autotool_make_shared
   name=fontconfig autotool_make_shared --with-arch=$gnu_arch --with-freetype-config=$ptlinsdk/$platform/usr/bin/freetype-config
   name=pixman autotool_make_shared
   name=cairo autotool_make_shared --enable-xlib --enable-xlib-xrender --disable-quartz --enable-png
   name=pango autotool_make_shared --enable-cairo --with-cairo
   name=gtk autotool_make_shared --without-libtiff --without-libjpeg

      # un-mangle include paths
   echo_cmd cd $platform/include
   echo_cmd cp ../lib/glib-2.0/include/glibconfig.h .
   echo_cmd cp ../lib/gtk-2.0/include/gdkconfig.h .
   echo_cmd ln -s -f pango-1.0/pango
   echo_cmd ln -s -f gtk-2.0/gtk
   echo_cmd ln -s -f gtk-2.0/gdk
   echo_cmd ln -s -f gtk-2.0/gdk-pixbuf
   echo_cmd ln -s -f gtk-2.0/gdk-pixbuf-xlib
   echo_echo1 "#include <cairo/cairo.h>" cairo.h
   echo_cmd ln -s -f glib-2.0/glib
   echo_cmd ln -s -f glib-2.0/gobject
   echo_cmd ln -s -f glib-2.0/gio
   echo_cmd ln -s -f glib-2.0/glib.h
   echo_cmd ln -s -f glib-2.0/glib-object.h
   echo_cmd ln -s -f glib-2.0/glib-unix.h
   echo_cmd ln -s -f glib-2.0/gmodule.h
   echo_cmd ln -s -f atk-1.0/atk
   echo_cmd cd $ptlinsdk
}

make_target_libs()
{
   # gnu_arch
   # tuple
   local platform="lin-$arch"
   local done=""

   name=ncurses autotool_make_shared --without-normal --includedir=$ptlinsdk/$platform/usr/include
   name=libxml2 autotool_make_shared --without-python
   unset CC CFLAGS
   name=zlib CHOST="$tuple" configure_make
   make_x11
   GTK_UPDATE_ICON_CACHE=gtk-update-icon-cache GLIB_COMPILE_SCHEMAS=glib-compile-schemas make_gtk
}

config_packages()
{
   tar_download atk
   tar_download binutils
   tar_download bison
   tar_download byacc
   tar_download bzip2
   tar_download cairo
   tar_download findutils
   tar_download flex
   tar_download fontconfig
   tar_download freetype
   tar_download gawk
   config_gcc build
   tar_download gcc
   config_gcc pass1
   tar_download gcc
   config_gcc pass2
   tar_download gcc
   config_gcc final
   tar_download gcc
   tar_download gettext
   tar_download glib
   tar_download glibc
   tar_download glibc_ports
   tar_download gmp
   tar_download gperf
   tar_download groff
   tar_download gtk
   tar_download inputproto
   tar_download kbproto
   tar_download libarchive
   tar_download libelf
   tar_download libffi
   tar_download libgcrypt
   tar_download libgpg_error
   tar_download libiconv
   tar_download libpng
   tar_download libpthread_stubs
   tar_download libtool
   tar_download libx11
   tar_download libxau
   tar_download libxcb
   tar_download libxext
   tar_download libxml2
   tar_download libxrender
   tar_download libxslt
   tar_download linux
   tar_download m4
   tar_download make
   tar_download mpc
   tar_download mpfr
   tar_download ncurses
   tar_download pango
   tar_download pixman
   tar_download pkg_config
   tar_download python
   tar_download renderproto
   tar_download sed
   tar_download xcb_proto
   tar_download xextproto
   tar_download xproto
   tar_download xtrans
   tar_download xz
   tar_download zlib
   if [ $MAKE_RHEL5_KERNELS ]; then
      tar_download rhel5
   fi
   if [ $MAKE_RHEL6_KERNELS ]; then
      tar_download rhel6
   fi
   if [ $MAKE_RHEL7_KERNELS ]; then
      tar_download rhel7
   fi
}

main()
{
   rm obj/all-done

      # Setup whitelisted system bin folders, and clear the path.
      # We want to make sure we know and control used system tools,
      # shells, utilities, so we can expect predicatable results
      # across different systems. This also prevents autoconf path
      # version searching nonsense. We want autoconf to predictably
      # use what we give it, not paw through the system paths
      # looking for things.
   if ! test -e build/sysbin; then
      echo_cmd rm -r -f build/sysbin
      echo_cmd mkdir -p build/sysbin
      echo_cmd rm -r -f build/systools
      echo_cmd mkdir -p build/systools
      for a in $(cat src/sysbin); do
         b="$(which $a)"
         if test $b; then
            if test -e src/stop/$a; then
               echo_cmd ln -s "$b" build/systools/$a
            else
               echo_cmd ln -s "$b" build/systools/$a
            fi
         fi
      done
   fi
   export PATH="$ptlinsdk/build/systools:$ptlinsdk/build/sysbin"

      # detect system and architecture
   common_cflags="-g -O2"
   build_mac_arch=""
   build_lin_arch=""
   build_cflags=""
   build_ldflags=""
   build_gcc_configure=""
   case $(uname) in
   (Linux)
      build_sys="lin"
      sys_path="$PATH"
      if [ -z "$MAKE_CONCURRENCY" ]; then
         MAKE_CONCURRENCY="$(($(grep -c -x "processor.*:.*" /proc/cpuinfo)+1))"
      fi
      case $(uname -m) in
      (x86_64)
         build_arch="x64"
         build_gnu_arch="x86_64"
         build_cflags="-march=athlon64"
         build_gcc_configur="--with-arch=athlon64"
         ;;
      (*86)
         build_arch="x86"
         build_gnu_arch="i686"
         build_lin_arch="i386"
         build_cflags="-march=i686 -mmmx -msse"
         build_gcc_configure="--with-arch=i686"
         ;;
      (powerpc64|ppc64)
         build_arch="ppc64"
         build_gnu_arch="powerpc64"
         # cpu-970 is g5
         build_cflags="-mcpu=970 -m64"
         ;;
      (powerpc*|ppc*)
         build_arch="ppc"
         build_gnu_arch="powerpc"
         # cpu-750 is g3
         build_cflags="-mcpu=750 -m32"
         ;;
      (arm*)
         build_arch="arm"
         build_gnu_arch="arm"
         ;;
      esac
      [ "$build_lin_arch" ] || build_lin_arch="$build_gnu_arch"
      build_tuple="$build_gnu_arch-build-linux-gnu"
      ;;
   (Darwin)
      build_sys="mac"
      sys_path="/Developer/usr/bin:$PATH"
      if [ -z "$MAKE_CONCURRENCY" ]; then
         MAKE_CONCURRENCY="$(($(sysctl hw.ncpu | sed "s/hw.ncpu: //")+1))"
      fi
      # export MACOSX_DEPLOYMENT_TARGET=10.7
      build_cflags="-Wno-unused-command-line-argument -mmacosx-version-min=10.7"
      build_ldflags="-macosx_version_min 10.7"
      if sysctl hw.cpu64bit_capable | grep -q ": 1"; then
         case $(uname -m) in
         (*86*)
            build_arch="x64"
            build_gnu_arch="x86_64"
            ;;
         (power*)
            build_arch="ppc64"
            build_gnu_arch="powerpc64"
            ;;
         esac
      else
         case $(uname -m) in
         (*86*)
            build_arch="x86"
            build_gnu_arch="i686"
            build_mac_arch="i386"
            build_lin_arch="i386"
            ;;
         (power*)
            build_arch="ppc"
            build_gnu_arch="powerpc"
            ;;
         (arm*)
            build_arch="arm"
            build_gnu_arch="arm"
            ;;
         esac
      fi
      [ "$build_mac_arch" ] || build_mac_arch="$build_gnu_arch"
      [ "$build_lin_arch" ] || build_lin_arch="$build_gnu_arch"
      build_tuple="$build_gnu_arch-build-darwin"
   esac
   build_platform="$build_sys-$build_arch"
   build_cflags="$build_cflags -isystem $ptlinsdk/build/usr/include -isystem /usr/include -L$ptlinsdk/build/usr/lib"
   build_ldflags="$build_ldflags -L$ptlinsdk/build/usr/lib -rpath $ptlinsdk/build/usr/lib"

   pwd
   test -d tars || echo_cmd mkdir -p tars
   test -d obj/part || echo_cmd mkdir -p obj/part

      # Download everything up front, so internet can be down
      # during most of the build, and so download issues do not
      # cause stoppage while unattended.
   for platform in build $MAKE_ARCH_LIST; do
      config_common
      config_$platform
      config_packages
   done

   platform="build"
   tuple="$build_tuple"
   gnu_arch="$build_gnu_arch"
   mac_arch="$build_mac_arch"
   lin_arch="$build_lin_arch"
   CONCURRENCY=$MAKE_CONCURRENCY
   setup_env
   setup_platform_folder
   setup_sys_libs

   config_common
   config_build
   config_packages

      # Bits needed to build our controlled gcc version are
      # compiled with the system compiler and binutils. These
      # builds can and will break as system tools and libraries
      # are upgraded. New patches will be needed over time, or
      # users will need to manually install back-rev system tools.
   name=m4 autotool_make_build
   name=make autotool_make_build
   name=gawk autotool_make_build
   name=sed autotool_make_build
   name=byacc autotool_make_build
   test -e build/bin/byacc || echo_cmd ln -s yacc build/bin/byacc
   name=flex autotool_make_build
   name=bison autotool_make_build
   name=gmp autotool_make_shared
   name=mpfr autotool_make_shared
   name=mpc autotool_make_shared
   name=libelf autotool_make_shared
   name=binutils autotool_make_build --target=$tuple --program-prefix=$tuple- --with-lib-path=$ptlinsdk/build/lib --program-suffix=- --enable-64-bit-bfd --with-sysroot --disable-nls --disable-multilib --disable-lto
   config_gcc build
   obj_name="$gcc_name" rpath="$ptlinsdk/build/lib" make_gcc --enable-languages=c,c++ $build_gcc_configure
   make_build_includes
   export PATH="$ptlinsdk/build/sysbin"

      # Remaining build tools compiled with our own known gcc
      # version. System perl, shell, system libs, and system
      # headers, are still used and may still break things
      # over time.
   name=libtool autotool_make_build --program-prefix=$tuple- --program-suffix=-
   name=pkg_config autotool_make_build --program-prefix=$tuple- --program-suffix=-
   name=groff autotool_make_build --without-x
   name=gperf autotool_make_build
   name=findutils autotool_make_build
   name=zlib configure_make_build
   name=bzip2 configure_make_build
   name=xz autotool_make_shared
   name=ncurses autotool_make_shared --without-normal --includedir=$ptlinsdk/$platform/usr/include
   name=libiconv autotool_make_shared
   name=libarchive autotool_make_shared
   name=python autotool_make_shared
   name=libxml2 autotool_make_shared
   name=libgpg_error autotool_make_shared
   name=libgcrypt autotool_make_shared
   name=libxslt autotool_make_shared
   name=xproto autotool_make_shared
   # to get msgfmt tool needed for target builds of gtk dependencies
   name=gettext autotool_make_shared
   # need all of gtk and depencies to get
   # build/bin/gtk-update-icon-cache for gtk target builds
   make_x11
   GTK_UPDATE_ICON_CACHE=./gtk-update-icon-cache make_gtk

   if is_arch_enabled x64 || test $build_arch = x64; then
      config_common
      config_x64
      config_packages
      arch="x64"
      gnu_arch="x86_64"
      lin_arch="x86_64"
      tuple="x86_64-pc-linux-gnu"
      target_cflags="-march=athlon64"
      gcc_configure="--with-arch=athlon64"
      glibc_configure=""
      make_gcc_toolchain
         gcc_ver=5.3.0
         gcc_url_="$gnumirror_url/gcc/gcc-$gcc_ver/"
         unset gcc_name
         config_packages
         obj_name="${gcc_name}_final"
         make_gcc --enable-__cxa_atexit --enable-languages=c,c++ $gcc_configure
      make_target_libs
   fi

   if is_arch_enabled x86 || test $build_arch = x86; then
      config_common
      config_x86
      config_packages
      arch="x86"
      gnu_arch="i686"
      lin_arch="i386"
      tuple="i686-pc-linux-gnu"
      target_cflags="-march=i686 -mmmx -msse"
      gcc_configure="--with-arch=i686"
      glibc_configure=""
      make_gcc_toolchain
      make_target_libs
   fi

   if is_arch_enabled ppc64 || test $build_arch = ppc64; then
      config_common
      config_ppc64
      config_packages
      arch="ppc64"
      gnu_arch="powerpc64"
      lin_arch="powerpc"
      tuple="powerpc64-pc-linux-gnu"
      target_cflags="-mcpu=970 -m64"
      gcc_configure="--with-cpu=970"
      glibc_configure=""
      make_gcc_toolchain
      make_target_libs
   fi

   if is_arch_enabled ppc || test $build_arch = ppc; then
      config_common
      config_ppc
      config_packages
      arch="ppc"
      gnu_arch="powerpc"
      lin_arch="powerpc"
      tuple="powerpc-pc-linux-gnu"
      # cpu-750 is g3
      target_cflags="-mcpu=750 -m32"
      gcc_configure="--with-cpu=750"
      glibc_configure=""
      make_gcc_toolchain
      make_target_libs
   fi

   if is_arch_enabled arm || test $build_arch = arm; then
      config_common
      config_arm
      config_packages
      arch="arm"
      gnu_arch="arm"
      lin_arch="arm"
      tuple="arm-pc-linux-gnueabi"
      target_cflags="-march=armv7 -mabi=aapcs-linux -mfloat-abi=soft -meabi=5"
      gcc_configure="--with-arch=armv5t --with-float=soft"
      glibc_configure="--without-fp"
      make_gcc_toolchain
      make_target_libs
   fi

   if is_arch_enabled x64; then
      config_common
      config_x64
      config_packages
      arch="x64"
      gnu_arch="x86_64"
      lin_arch="x86_64"
      tuple="x86_64-pc-linux-gnu"
      if [ $MAKE_RHEL5_KERNELS ]; then
         name=rhel5 config=      make_rhel
         name=rhel5 config=debug make_rhel
      fi
      if [ $MAKE_RHEL6_KERNELS ]; then
         name=rhel6 config=      make_rhel
         name=rhel6 config=debug make_rhel
      fi
      if [ $MAKE_RHEL7_KERNELS ]; then
         name=rhel7 config=      make_rhel
         name=rhel7 config=debug make_rhel
      fi
   fi

   config_common
   config_packages
   if [ $MAKE_RHEL5_KERNELS ]; then
      name=rhel5 strip_rhel
   fi
   if [ $MAKE_RHEL6_KERNELS ]; then
      name=rhel6 strip_rhel
   fi
   if [ $MAKE_RHEL7_KERNELS ]; then
      name=rhel7 strip_rhel
   fi

   echo_cmd touch obj/all-done
}


# start

# find self
arg0="$(which $0)"
[ "$arg0" ] && arg0="$0"
arg0_folder="$(dirname $arg0)"
arg0_folder="$(cd "$arg0_folder" && pwd)"
arg0_basename="$(basename $arg0)"
arg0="$arg0_folder/$arg0_basename"
ptlinsdk="$(cd $arg0_folder/.. && pwd)"
echo_cmd cd $ptlinsdk
cmp src/make.sh $arg0 || exit 1

mkdir -p obj
mv obj/make8.log obj/make9.log >/dev/null 2>&1
mv obj/make7.log obj/make8.log >/dev/null 2>&1
mv obj/make6.log obj/make7.log >/dev/null 2>&1
mv obj/make5.log obj/make6.log >/dev/null 2>&1
mv obj/make4.log obj/make5.log >/dev/null 2>&1
mv obj/make3.log obj/make4.log >/dev/null 2>&1
mv obj/make2.log obj/make3.log >/dev/null 2>&1
mv obj/make1.log obj/make2.log >/dev/null 2>&1
mv obj/make.log obj/make1.log >/dev/null 2>&1
mv obj/_make8.log obj/_make9.log >/dev/null 2>&1
mv obj/_make7.log obj/_make8.log >/dev/null 2>&1
mv obj/_make6.log obj/_make7.log >/dev/null 2>&1
mv obj/_make5.log obj/_make6.log >/dev/null 2>&1
mv obj/_make4.log obj/_make5.log >/dev/null 2>&1
mv obj/_make3.log obj/_make4.log >/dev/null 2>&1
mv obj/_make2.log obj/_make3.log >/dev/null 2>&1
mv obj/_make1.log obj/_make2.log >/dev/null 2>&1
mv obj/_make.log obj/_make1.log >/dev/null 2>&1
export ptlinsdk_log="$ptlinsdk/obj/_make.log"
rm -f obj/make.log $ptlinsdk_log >/dev/null 2>&1
touch $ptlinsdk_log >/dev/null 2>&1
main 2>&1 | tee obj/make.log
if test -f obj/all-done ; then
   echo
   echo --xbt-- completed
   echo
else
   echo
   echo --xbt-- error, something went wrong
   echo
fi
